//////////////////////////////////////////////////////////////////
//  TexturedMesh.cpp: implementation of the TexturedMesh class  //
//////////////////////////////////////////////////////////////////

#pragma warning( disable : 4786 )
#pragma warning( disable : 4503 )

#include "TexturedMesh.h"
#include "ColorTable.h"
#include <algorithm>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <set>
#include <map>
#include <deque>

//#include <GL/glut.h>
//#include <windows.h>


//////////////////////////////////////////////////////////////////////
// Constants
//////////////////////////////////////////////////////////////////////

const UINT TexturedMesh::INVALID_INDEX = -1; // in fact, this is maximal int...


//////////////////////////////////////////////////////////////////////
// I/O methods
//////////////////////////////////////////////////////////////////////


/**********************************************************************/
// WriteMeshToVRML: 
//
// Writes the mesh to out_file in VRML following format.
/**********************************************************************/
// WriteToVRML:
// Writes the mesh to a file, in VRML format.
// If write_charts is true, the mesh is written
// with triangles colors taken from m_triangles_colors.
// If write_charts is false, the mesh is written with
// texture.
// Returns true on success, false otherwise.
// geo_source: 0 for m_vertices
//             1 for m_exp_vertices (delta)
//			   2 for m_quant_vertices (xyz quantization)
// if 1 or 2 is not present writes m_vertices
bool TexturedMesh::WriteToVRML(ofstream &out_file) const
{
	char VRML_header[] = "#VRML V2.0 utf8";
	
	// Definition of VRML scene with defauls values
	char default_scene_begin[10000] = {'\0'}; 
	char texture_url[256] = {'\0'};
	
	if (m_textures.size() > 0)
		sprintf(texture_url, "texture ImageTexture {\n\t\turl \"%s\"\n}\n", m_textures[0]);
	
	sprintf(default_scene_begin, 
	"DEF Line01 Transform {\n \
								translation 39.91 -13.38 0 \n \
								rotation -1 0 0 -1.571 \n \
								children [ \n \
								Shape {    \n \
								appearance Appearance {  \n \
								material Material {  \n \
								diffuseColor 0.0 0.0 0.9   \n \
								ambientIntensity 0.4  \n \
								specularColor 0.9 0.9 0.9  \n \
								shininess 0.9  \n \
								transparency 0 \n \
								} \n \
								%s \n \
								} \n \
								geometry DEF Line01-FACES IndexedFaceSet { \n \
								ccw TRUE  \n \
								solid FALSE  \n", texture_url);
	
	
	out_file << VRML_header << endl << endl;
	out_file << default_scene_begin;
	
	out_file << "coord DEF Line01-COORD Coordinate { point [" << endl;
	
	// Printing the coordinates of the vertices.
	// We print 3 points in one row. The points are
	// separated by commas.
	const std::vector<Point3D> * geo = &m_vertices;
	for (UINT i = 0; i < (*geo).size() - 1 ; ++i)
	{
		if (i % 3 == 0)
			out_file << endl;
		
		out_file << (*geo)[i] << ", ";
	}
	out_file << (*geo)[(*geo).size() - 1] << "]" << endl;
	out_file << "\t}\n";
	
	
	// Printing texture coordinates if needed
	if (m_textures.size() > 0)
	{
		out_file << "       texCoord DEF Line01-TEXCOORD TextureCoordinate { point [\n";
		for (UINT u = 0; u < m_uv_coordinates.size() - 1; ++u)
		{
			out_file << m_uv_coordinates[u] << ", ";
			if (u+1 % 10 == 0)
				out_file << endl;
		}
		out_file << m_uv_coordinates[m_uv_coordinates.size() - 1] << "]" << endl;
		out_file << "\t}\n";
	}
	
	
	// Printing the faces (tripplets).
	// We print 3 tripplets in a row. They are
	// separated by -1.
	out_file << "\tcoordIndex [";
	for (UINT j = 0; j < m_faces.size() - 1; ++j)
	{
		if (j % 3 == 0)
			out_file << endl;
		
		out_file << m_faces[j] << "-1 ";
	}
	out_file << m_faces[m_faces.size() - 1] << "-1]\n" << endl;
	
	
	// Printing texture indices per triangle if needed
	if (m_textures.size() > 0)
	{
		out_file << "        texCoordIndex [\n";
		for (UINT tri = 0; tri < m_uv_per_triangle.size(); ++tri)
		{
			out_file << m_uv_per_triangle[tri] << " -1 ";
			if (tri+1 % 3 == 0)
				out_file << endl;
		}
		out_file << "]\n";
	}
	
	out_file << "\n}\n}" << endl;
	out_file << "\n]\n}" << endl;
	
	return (out_file.good() != 0);
}

/**********************************************************************/
// ReadMeshFromSimpleFormat: 
//
// Reads the mesh from in_file in the following format:
//
// num_vertices  num_triangles
// vertices     [list of 3D points]
// triangles    [list of tripplets]
/**********************************************************************/
void TexturedMesh::ReadFromSimpleFormat(ifstream & in_file)
{
	int num_vertices = 0;
	int num_triangles = 0;
	
	in_file >> num_vertices;
	in_file >> num_triangles;
	
	// Clean the existing structures of the mesh,
	// as we are reading new data
	m_vertices.clear();
	m_faces.clear();
	m_edge_list.clear();
	m_edge_to_triangles_map.clear();
	m_triangles_adjacency_list.clear();
	
	// Assigning space for the new data
	m_vertices.reserve(num_vertices);
	m_faces.reserve(num_triangles);
	
	Point3D cur_point;
	// Read points into m_vertices vector.	
	for (int v = 0; v < num_vertices; ++v) {
		in_file >> cur_point;
		m_vertices.push_back(cur_point);
	}
	
	Tripplet cur_tripplet;
	// Read triangles (tripplets of indices) into m_faces.
	for (int t = 0; t < num_triangles; ++t) {
		in_file >> cur_tripplet;
		m_faces.push_back(cur_tripplet);
	}
	
	if (IsMesh2D())
		m_is2Dmesh = true;
	else
		m_is2Dmesh = false;
}

// #vertices #faces
// vertices color indices
// normals
// faces
bool TexturedMesh::ReadFromPly4Format(ifstream & in_file) {
	int num_vertices = 0;
	int num_triangles = 0;
	
	in_file >> num_vertices;
	in_file >> num_triangles;
	
	// Clean the existing structures of the mesh,
	// as we are reading new data
	this->Clear();
		
	// Assigning space for the new data
	m_vertices.reserve(num_vertices);
	m_faces.reserve(num_triangles);
	m_own_vertex_normals.reserve(num_vertices);
	m_color_per_vertex.reserve(num_vertices);
	
	Point3D cur_point;
	int color_index;
	// Read points into m_vertices vector.	
	int v;
	for (v = 0; v < num_vertices; ++v) {
		in_file >> cur_point;
		m_vertices.push_back(cur_point);

		in_file >> color_index;
		const unsigned char * cur_color = GetColorFromTable(color_index);
		m_color_per_vertex.push_back(Point3D(cur_color[0]/255.0,cur_color[1]/255.0,cur_color[2]/255.0));
	}
	// Read normals into m_own_vertex_normals vector.	
	for (v = 0; v < num_vertices; ++v) {
		in_file >> cur_point;
		m_own_vertex_normals.push_back(cur_point);
	}
	
	Tripplet cur_tripplet;
	int supress_facenum;
	// Read triangles (tripplets of indices) into m_faces.
	for (int t = 0; t < num_triangles; ++t) {
		in_file >> supress_facenum;
		in_file >> cur_tripplet;
		m_faces.push_back(cur_tripplet);
	}
	
	if (IsMesh2D())
		m_is2Dmesh = true;
	else
		m_is2Dmesh = false;
	
	//  return (in_file.good() != 0);
	return true;
}


// #vertices #faces
// vertices
// normals
// faces
bool TexturedMesh::ReadFromPly3Format(ifstream & in_file) {
	int num_vertices = 0;
	int num_triangles = 0;
	
	in_file >> num_vertices;
	in_file >> num_triangles;
	
	// Clean the existing structures of the mesh,
	// as we are reading new data
	this->Clear();
		
	// Assigning space for the new data
	m_vertices.reserve(num_vertices);
	m_faces.reserve(num_triangles);
	m_own_vertex_normals.reserve(num_vertices);
	
	Point3D cur_point;
	// Read points into m_vertices vector.	
	int v;
	for (v = 0; v < num_vertices; ++v) {
		in_file >> cur_point;
		m_vertices.push_back(cur_point);
	}
	// Read normals into m_own_vertex_normals vector.	
	for (v = 0; v < num_vertices; ++v) {
		in_file >> cur_point;
		m_own_vertex_normals.push_back(cur_point);
	}
	
	Tripplet cur_tripplet;
	int supress_facenum;
	// Read triangles (tripplets of indices) into m_faces.
	for (int t = 0; t < num_triangles; ++t) {
		in_file >> supress_facenum;
		in_file >> cur_tripplet;
		m_faces.push_back(cur_tripplet);
	}
	
	if (IsMesh2D())
		m_is2Dmesh = true;
	else
		m_is2Dmesh = false;
	
	//  return (in_file.good() != 0);
	return true;
}


bool TexturedMesh::ReadFromPly2Format(ifstream & in_file) {
	int num_vertices = 0;
	int num_triangles = 0;
	
	in_file >> num_vertices;
	in_file >> num_triangles;
	
	// Clean the existing structures of the mesh,
	// as we are reading new data
	m_vertices.clear();
	m_faces.clear();
	m_edge_list.clear();
	m_edge_to_triangles_map.clear();
	m_triangles_adjacency_list.clear();
	
	// Assigning space for the new data
	m_vertices.reserve(num_vertices);
	m_faces.reserve(num_triangles);
	
	Point3D cur_point;
	// Read points into m_vertices vector.	
	for (int v = 0; v < num_vertices; ++v) {
		in_file >> cur_point;
		m_vertices.push_back(cur_point);
	}
	
	Tripplet cur_tripplet;
	int supress_facenum;
	// Read triangles (tripplets of indices) into m_faces.
	for (int t = 0; t < num_triangles; ++t) {
		in_file >> supress_facenum;
		in_file >> cur_tripplet;
		m_faces.push_back(cur_tripplet);
	}
	
	if (IsMesh2D())
		m_is2Dmesh = true;
	else
		m_is2Dmesh = false;
	
	//  return (in_file.good() != 0);
	return true;
}

bool TexturedMesh::WriteToPly2Format(ofstream &out_file) {
	Point3D * geoArray = &(m_vertices[0]);
	
	out_file << GetNumberOfVertices();
	out_file << " " << GetNumberOfTriangles() << endl;
	
	for (int v = 0; v < GetNumberOfVertices(); ++v) {
		out_file << geoArray[v] << endl;
	}
	
	for (int t = 0; t < GetNumberOfTriangles(); ++t) {
		out_file << "3 " << m_faces[t].GetFirstIndex() << " " << m_faces[t].GetSecondIndex() << " " << m_faces[t].GetThirdIndex() << endl;
	}
	
	return (out_file.good() != 0);
}


/**********************************************************************/
// GetBoundingBox:
//
// Returns axes-aligned bounding box of the mesh.
/**********************************************************************/
Box TexturedMesh::GetBoundingBox() const {
	// If the mesh is empty return a null bb
	if (m_faces.size() <= 0)
		return Box(0, 0, 0, 0, 0, 0);
	
	real minX(REAL_MAX), minY(REAL_MAX), minZ(REAL_MAX), 
		maxX(REAL_MIN), maxY(REAL_MIN), maxZ(REAL_MIN);
	
	const std::vector<Point3D> & geometry = m_vertices;
	
	for (UINT i = 0; i < geometry.size(); ++i) {
		const Point3D & vertex = geometry[i];
		minX = min(minX, vertex.GetX());
		minY = min(minY, vertex.GetY());
		minZ = min(minZ, vertex.GetZ());
		maxX = max(maxX, vertex.GetX());
		maxY = max(maxY, vertex.GetY());
		maxZ = max(maxZ, vertex.GetZ());
	}
	
	return Box(minX, maxX, minY, maxY, minZ, maxZ);
}

//////////////////////////////////////////////////////////////////////
// Auxiliary (private) functions
//////////////////////////////////////////////////////////////////////

/******************************************************************/
// CreateEdgeList:                                                
// 
// Constructs an adjacency list according to the faces data.
// The function assumes that m_vertices
// and m_faces are already initialized and correct.
/******************************************************************/
void TexturedMesh::CreateEdgeList()
{
	// There are no vertices - no work to be done
	if (m_vertices.size() == 0)
		return;
	
	// Assign space for the edge vector
	m_edge_list.resize(m_vertices.size());
	
	Tripplet cur_face;
	for (UINT i = 0; i < m_faces.size(); ++i) {
		cur_face = m_faces[i];
		
		// Each triangle contributes 3 (undirectional) edges
		ScanAndAdd(m_edge_list[cur_face.GetFirstIndex()], cur_face.GetSecondIndex());
		ScanAndAdd(m_edge_list[cur_face.GetSecondIndex()], cur_face.GetFirstIndex());
		
		ScanAndAdd(m_edge_list[cur_face.GetFirstIndex()], cur_face.GetThirdIndex());
		ScanAndAdd(m_edge_list[cur_face.GetThirdIndex()], cur_face.GetFirstIndex());
		
		ScanAndAdd(m_edge_list[cur_face.GetSecondIndex()], cur_face.GetThirdIndex());
		ScanAndAdd(m_edge_list[cur_face.GetThirdIndex()], cur_face.GetSecondIndex());
	}
}



/******************************************************************/
// ScanAndAdd:                                                
// 
// Add index to the end of ind_vector if it's not already there
/******************************************************************/
void TexturedMesh::ScanAndAdd(IndexVector & ind_vector, UINT index) {
	if (std::find(ind_vector.begin(), ind_vector.end(), index) == ind_vector.end())
		ind_vector.push_back(index);
}


/******************************************************************/
// CreateEdgeToTrianglesMap:                                                
// 
// This function builds m_edge_to_triangles_map. It assumes
// that m_vertices, m_faces and m_edge_list are initialized correctly.
//
// Algorithm by Ronen Gvili - runs in O(N logN) where N is number
// of triangles in the mesh.
/******************************************************************/
void TexturedMesh::CreateEdgeToTrianglesMap() {
	Tripplet cur_tripplet;
	
	// Go over the triangles. Each triangle has 3 edges,
	// so we add the triangle's index to the 3 edge entries in the map.
	for (UINT tri_ind = 0; tri_ind < m_faces.size(); ++tri_ind) {
		cur_tripplet = m_faces[tri_ind];
		
		// Insert each pair of indices (=edge) into the map,
		// and add tri_ind as the value.
		AddIndexToEdgeMap(m_edge_to_triangles_map, cur_tripplet.GetFirstIndex(),
			cur_tripplet.GetSecondIndex(), tri_ind);
		AddIndexToEdgeMap(m_edge_to_triangles_map, cur_tripplet.GetFirstIndex(),
			cur_tripplet.GetThirdIndex(), tri_ind);
		AddIndexToEdgeMap(m_edge_to_triangles_map, cur_tripplet.GetSecondIndex(),
			cur_tripplet.GetThirdIndex(), tri_ind);
	}
}



/******************************************************************/
// CreateTrianglesAdjacencyList:                                                
// 
// This function builds m_triangles_adjacency_list. It assumes
// that m_vertices, m_faces, m_edge_list and m_edge_to_triangles_map
// are initialized correctly.
//
// Algorithm by Ronen Gvili - runs in O(N logN) where N is number
// of triangles in the mesh.
/******************************************************************/
void TexturedMesh::CreateTrianglesAdjacencyList() {
	if (m_faces.size() == 0) // no triangles - no work to be done
		return;
	
	if (m_edge_list.empty())
		CreateEdgeList();
	
	if (m_edge_to_triangles_map.empty())
		CreateEdgeToTrianglesMap();
	
	// Assign space in the vector
	m_triangles_adjacency_list.resize(m_faces.size());	
	
	// Go over the edge-to-triangles map. Each edge entry has
	// vector of its triangles i.e. triangles that lie on it).
	// If two triangles lie on an edge, then they are neighbours!
	// We assume that the mesh is legal - no more than 2 triangles
	// on one edge.
	EdgeToIndicesMap::iterator edge_iter = m_edge_to_triangles_map.begin();
	for ( ; edge_iter != m_edge_to_triangles_map.end(); ++edge_iter) {
		if ( (*edge_iter).second.size() > 1 ) { // the current edge has more 
			// than one triangle on it
			UINT first_tri_ind = (*edge_iter).second[0];
			UINT second_tri_ind = (*edge_iter).second[1];
			
			// first_tri_ind and second_tri_ind are neighbours,
			// so we add them to the adjacency list
			m_triangles_adjacency_list[first_tri_ind].push_back(second_tri_ind);
			m_triangles_adjacency_list[second_tri_ind].push_back(first_tri_ind);
		}
	}
}


/******************************************************************/
// AddIndexToEdgeMap:                                                
// 
// Adds triangle index ind to the vector 
// under key (vertex_ind1,vertex_ind2) in the edge_map.
/******************************************************************/
void TexturedMesh::AddIndexToEdgeMap(EdgeToIndicesMap & edge_map, 
									 UINT vertex_ind1, 
									 UINT vertex_ind2, 
									 UINT ind) {
	assert(vertex_ind1 != vertex_ind2); // the two vertices of an edge must be different
	
	// We sort the indices of edges in incresing order.
	// I.e. edge (i,j) is stored in the map so that (i<j).
	UINT min_ind = min(vertex_ind1, vertex_ind2);
	UINT max_ind = max(vertex_ind1, vertex_ind2);
	
	ScanAndAdd(edge_map[ IndexPair(min_ind, max_ind) ], ind);
}

/******************************************************************/
// CreateVertexToTrianglesMap:                                               
// 
// This function builds m_vertex_to_triangles_map. It scans
// the list of faces, adding the current triangle index
// to each of its vertices entries.
// The function assumes m_vertices and m_faces are initialized.
/******************************************************************/
void TexturedMesh::CreateVertexToTrianglesMap() {
	m_vertex_to_triangles_map.resize(m_vertices.size());
	
	for (UINT tri_ind = 0; tri_ind < m_faces.size(); ++tri_ind) {
		const Tripplet & cur_tripplet = m_faces[tri_ind];
		m_vertex_to_triangles_map[cur_tripplet.GetFirstIndex()].push_back(tri_ind);
		m_vertex_to_triangles_map[cur_tripplet.GetSecondIndex()].push_back(tri_ind);
		m_vertex_to_triangles_map[cur_tripplet.GetThirdIndex()].push_back(tri_ind);
	}
}


/******************************************************************/
// IsMesh2D: 
//
// Returns true if the mesh is flat
// against the Z axis (all vertices have z=0).
// Returns false if mesh isn't flat or if it's empty.
// Should be called after any alterations of the mesh.
/******************************************************************/
bool TexturedMesh::IsMesh2D() const {
	if (m_vertices.size() == 0)
		return false;
	
	for (UINT i = 0; i < m_vertices.size(); ++i) {
		if (m_vertices[i].GetZ() != real(0) )
			return false;
	}
	
	return true;
}


/******************************************************************/
// GetCenterOfMass: 
//
// Returns the center point
// of the mesh (the average of all vertices).
/******************************************************************/
Point3D TexturedMesh::GetCenterOfMass() const {
	const std::vector<Point3D> & geometry = m_vertices;
	Point3D sum(0,0,0);
	
	for (UINT i = 0; i < GetNumberOfVertices(); ++i)
		sum += geometry[i];
	
	return (1.0 / GetNumberOfVertices()) * sum;
}

/******************************************************************/
// Translate: 
//
// Translates the mesh by offset_vector
/******************************************************************/
void TexturedMesh::Translate(const Point3D & offset_vector) {
	for (UINT i = 0; i < GetNumberOfVertices(); ++i)
		m_vertices[i] += offset_vector;
}


/******************************************************************/
// Rotate2DMeshAroundOrigin: rotates 2D mesh
// by angle around the origin (0,0,0).
// If mesh is 3D, leaves as is.
/******************************************************************/
void TexturedMesh::Rotate2DMeshAroundOrigin(real angle) {
	
	if (! m_is2Dmesh)
		return;
	
	
	for (UINT i = 0; i < GetNumberOfVertices(); ++i)
		m_vertices[i].Rotate(angle);
}

/******************************************************************/
// Rotate2DMeshToBestBB: 
// If mesh is 2D, rotates it so its bounding
// rectangle is the tightest possible.
// If mesh is not 2D, leaves it as is.
/******************************************************************/
void TexturedMesh::Rotate2DMeshToBestBB()
{
	
	if (! m_is2Dmesh)
		return;
	
	// n - number of trial rotations
	int			n = 20;
	std::vector<real>	areaVec(n, 0);
	Point3D		center = GetCenterOfMass();
	
	Translate(Point3D(-1 * center.GetX(),
		-1 * center.GetY(),
		0));
	
	for (USHORT i = 0; i < n; ++i) {
		Rotate2DMeshAroundOrigin(PI /(2.0 * n));
		areaVec[i] = GetBoundingBox().GetBaseArea();
	}
	
	
	USHORT minInd = 0;
	real	minArea = REAL_MAX;
	for (i = 0; i < n; ++i) {
		if (areaVec[i] < minArea) {
			minInd = i;
			minArea = areaVec[i];
		}
	}
	
	Rotate2DMeshAroundOrigin(-1.0 * PI / 2.0);
	Rotate2DMeshAroundOrigin((minInd+1)*PI/(2.0*n));
	
	Translate(Point3D(center.GetX(),
		center.GetY(),
		0));
	
	Box bb = GetBoundingBox();
	Translate(Point3D(-1 * bb.GetLeft(), 
		-1 * bb.GetBack(),
		0));
}

void TexturedMesh::Scale(real factor) {
	Point3D center = GetCenterOfMass();
	
	for (UINT i = 0; i < m_vertices.size(); ++i)
		m_vertices[i] = factor*m_vertices[i] + (1-factor)*center;
}


/******************************************************************/
// CleanDegenerateTripplets: 
//
// Removes tripplets
// from m_faces (and accordingly from m_uv_per_triangle)
// that have two ore more identical indices.
// Should be called BEFORE building all edge/triangle maps!!
/******************************************************************/
void TexturedMesh::CleanDegenerateTripplets() {
	int i = 0;
	while (true) {
		if (i >= m_faces.size())
			break;
		
		Tripplet & cur_tripplet = m_faces[i];
		
		if ( (m_vertices[cur_tripplet.GetFirstIndex()] == m_vertices[cur_tripplet.GetSecondIndex()]) ||
			(m_vertices[cur_tripplet.GetFirstIndex()] == m_vertices[cur_tripplet.GetThirdIndex()]) ||
			(m_vertices[cur_tripplet.GetThirdIndex()] == m_vertices[cur_tripplet.GetSecondIndex()]) ) {
			m_faces[i] = m_faces[m_faces.size() - 1];
			m_faces.erase(m_faces.begin() + (m_faces.size() - 1));
			
			if (m_uv_per_triangle.size() > 0)	{
				m_uv_per_triangle[i] = m_uv_per_triangle[m_uv_per_triangle.size() - 1];
				m_uv_per_triangle.erase(m_uv_per_triangle.begin() + (m_uv_per_triangle.size() - 1));	
			}
		}
		else {
			++i;
		}
	}
}


real TexturedMesh::GetSurfaceArea() const {
	real ret = 0;
	for (UINT i = 0; i < m_faces.size(); ++i)
		ret += this->GetTriangle(i).GetArea();
	
	return ret;
}

/******************************************************************/
// WriteToSimpleFormat
/******************************************************************/
bool TexturedMesh::WriteToSimpleFormat(ofstream &out_file) const {
	const Point3D * geoArray = &(m_vertices[0]);
	
	out_file << GetNumberOfVertices();
	out_file << " " << GetNumberOfTriangles() << endl;
	
	for (int v = 0; v < GetNumberOfVertices(); ++v) {
		out_file << geoArray[v] << endl;
	}
	
	for (int t = 0; t < GetNumberOfTriangles(); ++t) {
		out_file << m_faces[t].GetFirstIndex() << " " << m_faces[t].GetSecondIndex() << " " << m_faces[t].GetThirdIndex() << endl;
	}
	
	
	return (out_file.good() != 0);
}

real TexturedMesh::GetTotalEdgeLength() {
	real len = 0;
	
	if (m_edge_to_triangles_map.size() == 0)
		CreateEdgeToTrianglesMap();
	
	EdgeToIndicesMap::const_iterator it = m_edge_to_triangles_map.begin();
	for ( ; it != m_edge_to_triangles_map.end(); ++it) {
		len += Point3D::Distance(m_vertices[(it->first).first],
			m_vertices[(it->first).second]);
	}
	
	return len;
}


/******************************************************************/
// GetMinEdgeLength:
//
// Returns minimal edge length
/******************************************************************/
real TexturedMesh::GetMinEdgeLength() {
	if (m_edge_to_triangles_map.size() == 0)
		CreateEdgeToTrianglesMap();
	
	EdgeToIndicesMap::const_iterator it = m_edge_to_triangles_map.begin();
	real min_len = Point3D::Distance(m_vertices[(it->first).first],
		m_vertices[(it->first).second]);
	real len1;
	for ( ; it != m_edge_to_triangles_map.end(); ++it) {
		len1 = Point3D::Distance(m_vertices[(it->first).first],
			m_vertices[(it->first).second]);
		
		min_len = (len1 < min_len)? len1 : min_len;
	}
	
	return min_len;
}


/******************************************************************/
// GetAverageEdgeLength:
//
// Returns average edge length
/******************************************************************/
real TexturedMesh::GetAverageEdgeLength() {
	return (this->GetTotalEdgeLength())/m_edge_to_triangles_map.size();
}



/******************************************************************/
// IsBoundaryVertex:
//
// Returns true if i is a vertex on mesh boundary
/******************************************************************/
bool TexturedMesh::IsBoundaryVertex(UINT i) {
	// Initializing mesh auxiliary structures
	if (m_edge_list.empty())
		CreateEdgeList();
	if (m_edge_to_triangles_map.empty())
		CreateEdgeToTrianglesMap();
	if (m_triangles_adjacency_list.empty())
		CreateTrianglesAdjacencyList();
	if (m_vertex_to_triangles_map.size() == 0)
		this->CreateVertexToTrianglesMap();
	
	for (int t = 0; t < m_vertex_to_triangles_map[i].size(); ++t)	{
		if (m_triangles_adjacency_list[ m_vertex_to_triangles_map[i][t] ].size() == 2) {
			Tripplet & t1 = m_faces[ m_triangles_adjacency_list[ m_vertex_to_triangles_map[i][t] ][0] ];
			Tripplet & t2 = m_faces[ m_triangles_adjacency_list[ m_vertex_to_triangles_map[i][t] ][1] ];
			if (Tripplet::IndexSetConjunction(t1, t2) != Tripplet::INVALID_INDEX) {
				if (i != Tripplet::IndexSetConjunction(t1, t2))
					return true;
			}
		}
		else if (m_triangles_adjacency_list[ m_vertex_to_triangles_map[i][t] ].size() == 1)
			return true;
	}
	
	return false;
}

/******************************************************************/
// IsIsolatedVertex:
//
// Returns true if i is an isolated vertex (has no neighbours)
/******************************************************************/
bool TexturedMesh::IsIsolatedVertex(UINT i) {
	// Initializing mesh auxiliary structures
	if (m_edge_list.empty())
		CreateEdgeList();
	
	if (m_edge_list[i].size() == 0)
		return true;
	else
		return false;
}

/******************************************************************/
// IsHangingVertex:
//
// Returns true if i is a hanging vertex (has only one neighbour)
/******************************************************************/
bool TexturedMesh::IsHangingVertex(UINT i) {
	// Initializing mesh auxiliary structures
	if (m_edge_list.empty())
		CreateEdgeList();
	
	if (m_edge_list[i].size() == 1)
		return true;
	else
		return false;
}


int TexturedMesh::Neighbours(UINT i, UINT j) {
	for (UINT k = 0; k < m_edge_list[i].size(); ++k) {
		if (m_edge_list[i][k] == j)
			return k;
	}
	
	return -1;
}

// returns true if (i,j) is an edge in the mesh graph
// false otherwise
bool TexturedMesh::Are1neighbours(UINT i, UINT j) {
	return (std::find(m_edge_list[i].begin(), m_edge_list[i].end(), j) != m_edge_list[i].end()); 
}



/******************************************************************/
// CleanIsolatedVertices: 
//
// Removes vertices that aren't
// connected to any other vertex. Also updates the m_faces
// vector accordingly. Destroys all the maps that are
// affected by this!!!!!
/******************************************************************/
void TexturedMesh::CleanIsolatedVertices() {
	std::vector<UINT> invalid(m_vertices.size(), 0);
	std::vector<UINT> map(m_vertices.size(), -1);
	
	
	bool flag = false;
	for (UINT i = 0; i < m_vertices.size(); ++i) {
		if (IsIsolatedVertex(i)) {
			invalid[i] = 1;
			flag = true;
		}
	}
	
	if (! flag) // no isolated vertices found
		return;
	
	// making new vertex array:
	std::vector<Point3D> new_vertex_array;
	
	i = 0;
	for (UINT j = 0; j < invalid.size(); ++j) {
		if (! invalid[j]) {
			new_vertex_array.push_back(m_vertices[j]);
			map[j] = i;
			++i;
		}
	}
	
	m_vertices.clear();
	m_vertices = new_vertex_array;
	
	new_vertex_array.clear();
	
	// fixing the topology
	for (i = 0; i < m_faces.size(); ++i) {
		Tripplet & tr = m_faces[i];
		UINT first  = tr.GetFirstIndex();
		UINT second = tr.GetSecondIndex();
		UINT third  = tr.GetThirdIndex();
		
		tr.SetFirstIndex(map[first]);
		tr.SetSecondIndex(map[second]);
		tr.SetThirdIndex(map[third]);
	}
	
	// destroying the irrelevant maps...
	m_edge_list.clear();
	m_vertex_to_triangles_map.clear();
	m_edge_to_triangles_map.clear();
}

void TexturedMesh::ReverseAllNormals() {
	int temp;
	
	for (UINT i = 0; i < m_faces.size(); ++i) {
		temp = m_faces[i].GetFirstIndex();
		m_faces[i].SetFirstIndex(m_faces[i].GetSecondIndex());
		m_faces[i].SetSecondIndex(temp);
	}
}

void TexturedMesh::SetMesh2Render() {
	// Initializing mesh auxiliary structures
	if (m_edge_list.empty())
		CreateEdgeList();
	if (m_edge_to_triangles_map.empty())
		CreateEdgeToTrianglesMap();
	if (m_triangles_adjacency_list.empty())
		CreateTrianglesAdjacencyList();
	if (m_vertex_to_triangles_map.size() == 0)
		this->CreateVertexToTrianglesMap();
	if (m_color_per_vertex.size() == 0)
		m_color_per_vertex = std::vector<Point3D>(m_vertices.size(), Point3D(1.0, 1.0, 1.0));
	if (m_color_per_vertex1.size() == 0)
		m_color_per_vertex1 = std::vector<Point3D>(m_vertices.size(), Point3D(1.0, 1.0, 1.0));
	
	
	// calculating normals:
	if (m_vertices.size() > 0) {
		m_vertex_normals.resize(m_vertices.size());
	}
	
	if (m_faces.size() > 0) {
		m_face_normals.resize(m_faces.size());
	}
	
	for (int i = 0; i < m_faces.size(); ++i) {
		m_face_normals[i] = Point3D::VectorProduct(m_vertices[ m_faces[i].GetSecondIndex() ] - 
					       m_vertices[ m_faces[i].GetFirstIndex() ],
						   m_vertices[ m_faces[i].GetThirdIndex() ] -
						   m_vertices[ m_faces[i].GetFirstIndex() ]);
		m_face_normals[i].Normalize();
	}
	
	for (i = 0; i < m_vertices.size(); ++i) {
		m_vertex_normals[i] = Point3D(0.0f, 0.0f, 0.0f);
		for (int j = 0; j < m_vertex_to_triangles_map[i].size(); ++j) {
			m_vertex_normals[i] += m_face_normals[ m_vertex_to_triangles_map[i][j] ];
		}
		m_vertex_normals[i].Normalize();
	}
}


// VertexDegree: returns the valency of vertex no.v
// returns -1 if index v is out of range
int TexturedMesh::VertexDegree(UINT v) {
	if (m_edge_list.empty())
		CreateEdgeList();
	
	if (v >= this->GetNumberOfVertices())
		return -1;
	
	return m_edge_list[v].size();
}

//void myGlutDisplay( void );

void TexturedMesh::FixOrientation() {
	if (m_triangles_adjacency_list.empty())
		CreateTrianglesAdjacencyList();
	
	std::deque<UINT> triQ;
	std::vector<bool> visitedTriangles(GetNumberOfTriangles(), false);
	
	triQ.push_back(0);
	visitedTriangles[0] = true;
	
	// BFS over the dual graph
	UINT currentTri;
	while (! triQ.empty() ) {
		currentTri = triQ.front();
		triQ.pop_front();
		
		for (IndexVector::const_iterator ntit = m_triangles_adjacency_list[currentTri].begin();
		ntit != m_triangles_adjacency_list[currentTri].end();
		++ntit) {
			if (! visitedTriangles[*ntit]) {
				if (Tripplet::AreReversed(m_faces[currentTri], m_faces[*ntit])) {
					m_faces[*ntit].Reverse();			
				}
				
				triQ.push_back(*ntit);
				visitedTriangles[*ntit] = true;
			}
		}
	}
}


// The file format used by the Metro tool
void TexturedMesh::WriteToSMFFormat(ofstream &out_file) const {
	const Point3D * geoArray = &(m_vertices[0]);
	
	
	// Header:
	out_file << "#$SMF 1.0" << endl;
	out_file << "#$vertices " << GetNumberOfVertices() << endl;
	out_file << "#$faces " << GetNumberOfTriangles() << endl;
	
	// Vertices
	for (int i=0; i < GetNumberOfVertices(); ++i) {
		out_file<<"v "<<geoArray[i].GetX()<<" "<<geoArray[i].GetY()<<" "<<geoArray[i].GetZ()<<endl;
	}
	
	// Faces : count from 1!!
	for (i=0; i < GetNumberOfTriangles(); ++i) {
		out_file<<"f "<<m_faces[i].GetFirstIndex()+1<<" "<<m_faces[i].GetSecondIndex()+1<<" "<<m_faces[i].GetThirdIndex()+1<<endl;
	}
}
