// Tripplet.h: interface for the Tripplet class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_TRIPPLET_H__9283BB44_5348_11D4_AE4D_00609716AB58__INCLUDED_)
#define AFX_TRIPPLET_H__9283BB44_5348_11D4_AE4D_00609716AB58__INCLUDED_

#include "defs.h"
#include "Serializable.h"

/** This class represents a 3D triangle composed of 3 vertices .
* represented by their indices in a triangular mesh.
*/
class Tripplet //: public Serializable
{
public:
	///
	static const ULONG INVALID_INDEX;
    ///
    Tripplet() throw();
	///
	Tripplet(ULONG index1, ULONG index2, ULONG index3) throw();
	///
	Tripplet(IN const Tripplet & tripplet) throw();
	///
	Tripplet & operator=(IN const Tripplet & tripplet) throw();
	///
	friend bool	operator==(IN const Tripplet & lhs, IN const Tripplet & rhs) throw();
	///
	friend bool	operator!=(IN const Tripplet & lhs, IN const Tripplet & rhs) throw();
	
	///
	ULONG GetFirstIndex() const throw();
	///
	ULONG GetSecondIndex() const throw();
	///
	ULONG GetThirdIndex() const throw();
	///
	ULONG GetIndex(USHORT i) const throw();

	///
	void  SetFirstIndex(ULONG val) throw();
	///
	void  SetSecondIndex(ULONG val) throw();
	///
	void  SetThirdIndex(ULONG val) throw();
	///
	void  SetIndex(USHORT i, ULONG val) throw();
	void Reverse() throw();
	///
	void RollLeft() throw();
	///
	void RollRight() throw();
	/// Premuts the Tripplet indices.
	void Permute(USHORT index0, USHORT index1) throw();
	///
	void AddOffset(ULONG offset) throw();

	// Returns the order index of val if val
	// is present in the tripplet, or 4 otherwise.
	USHORT FindIndex(ULONG val) const throw();

	/** Returns the difference between the sets of neighbouring
	* tripplet indices {lhs} \ {rhs} or INVALID_INDEX if the set is empty.
	* pre-condition: the size of the difference set is at most 1.
	*/
	static ULONG IndexSetDifference(const Tripplet & lhs,
		const Tripplet & rhs) throw();

	/** Returns the conjunction between the sets of neighbouring
	* tripplet indices {lhs} ^ {rhs} or INVALID_INDEX if the set is empty.
	* pre-condition: the size of the difference set is at most 1.
	*/
	static ULONG IndexSetConjunction(const Tripplet & lhs,
		const Tripplet & rhs) throw();

	/** Assumes the tripplets represent neighbouring triangles,
	* meaning that the disjunction of the two indices sets is of size 2.
	* Returns true if the two triangles are differently oriented
	*/
	static bool AreReversed(const Tripplet & lhs, const Tripplet & rhs) throw();


   	/** Serializable interface overrides */
	/// Reads a point from the stream.
	void Read(ifstream & stream);
	/// Writes a point to the stream.
	void Write(ofstream & stream) const;

	/// serializes an object to a stream
	friend ofstream & operator << (ofstream & stream, const Tripplet & object)
	{
		object.Write(stream);
		return stream;
	}
	
	/// serializes an object from a stream
	friend ifstream & operator >> (ifstream & stream, Tripplet & object)
	{
		object.Read(stream);
		return stream;
	}


private:
	///
	UINT m_indices[3];
};

//////////////////////////////////////////////////////////////////////
// Inline Implementations
//////////////////////////////////////////////////////////////////////

inline Tripplet::Tripplet() throw()
{ 
	m_indices[0]=0; 
	m_indices[1]=1; 
	m_indices[2]=2;
}

inline Tripplet::Tripplet(ULONG index1, ULONG index2, ULONG index3) throw()
{
	// Check that this is the invalid tripplet (0,0,0)
	// Otherwise all the indices should be different if they compose a triangle
	assert( (index1 == 0 && index2 == 0 && index3 == 0) ||
		(index1 != index2 && index1 != index3 && index2 != index3));

	m_indices[0] = index1;
	m_indices[1] = index2;
	m_indices[2] = index3;
}

inline Tripplet & Tripplet::operator=(IN const Tripplet & rhs) throw()
{
	m_indices[0] = rhs.m_indices[0];
	m_indices[1] = rhs.m_indices[1];
	m_indices[2] = rhs.m_indices[2];
	return *this;
}

inline Tripplet::Tripplet(IN const Tripplet & rhs) throw()
{
	m_indices[0] = rhs.m_indices[0];
	m_indices[1] = rhs.m_indices[1];
	m_indices[2] = rhs.m_indices[2];
}
	
inline ULONG Tripplet::GetFirstIndex() const throw()
{
	return m_indices[0];
}

inline ULONG Tripplet::GetSecondIndex() const throw()
{
	return m_indices[1];
}

inline ULONG Tripplet::GetThirdIndex() const throw()
{
	return m_indices[2];
}

inline ULONG Tripplet::GetIndex(USHORT i) const throw()
{
	assert(i < 3);
	return m_indices[i];
}

inline bool	operator==(IN const Tripplet & lhs, IN const Tripplet & rhs) throw()
{
	return ( (lhs.m_indices[0] == rhs.m_indices[0]) &&
		     (lhs.m_indices[1] == rhs.m_indices[1]) &&
		     (lhs.m_indices[2] == rhs.m_indices[2]) );
}

inline bool	operator!=(IN const Tripplet & lhs, IN const Tripplet & rhs) throw()
{
	return !(lhs == rhs);
}

inline ULONG Tripplet::IndexSetDifference(const Tripplet & lhs,
		const Tripplet & rhs) throw()
{
	ULONG differentIndex;

	if (lhs.GetFirstIndex() != rhs.GetFirstIndex() &&
		lhs.GetFirstIndex() != rhs.GetSecondIndex() &&
		lhs.GetFirstIndex() != rhs.GetThirdIndex())
	{
		differentIndex = lhs.GetFirstIndex();
	}
	else if (lhs.GetSecondIndex() != rhs.GetFirstIndex() &&
		lhs.GetSecondIndex() != rhs.GetSecondIndex() &&
		lhs.GetSecondIndex() != rhs.GetThirdIndex())
	{
		differentIndex = lhs.GetSecondIndex();
	}
	else if (lhs.GetThirdIndex() != rhs.GetFirstIndex() &&
		lhs.GetThirdIndex() != rhs.GetSecondIndex() &&
		lhs.GetThirdIndex() != rhs.GetThirdIndex())
	{
		differentIndex = lhs.GetThirdIndex();
	}
	else
	{
		return INVALID_INDEX;
	}
	return differentIndex;
}

inline ULONG Tripplet::IndexSetConjunction(const Tripplet & lhs,
	const Tripplet & rhs) throw()
{
	ULONG differentIndex;
	int count = 0;

	if (lhs.GetFirstIndex() == rhs.GetFirstIndex() ||
		lhs.GetFirstIndex() == rhs.GetSecondIndex() ||
		lhs.GetFirstIndex() == rhs.GetThirdIndex())
	{
		differentIndex = lhs.GetFirstIndex();
		count++;
	}
	if (lhs.GetSecondIndex() == rhs.GetFirstIndex() ||
		lhs.GetSecondIndex() == rhs.GetSecondIndex() ||
		lhs.GetSecondIndex() == rhs.GetThirdIndex())
	{
		differentIndex = lhs.GetSecondIndex();
		count++;
	}
	if (lhs.GetThirdIndex() == rhs.GetFirstIndex() ||
		lhs.GetThirdIndex() == rhs.GetSecondIndex() ||
		lhs.GetThirdIndex() == rhs.GetThirdIndex())
	{
		differentIndex = lhs.GetThirdIndex();
		count++;
	}
	if (count == 0 || count > 1)
	{
		return INVALID_INDEX;
	}
	return differentIndex;
}


inline void Tripplet::Reverse() throw()
{
	std::swap(m_indices[0], m_indices[2]);
}

inline void Tripplet::RollLeft() throw()
{
	std::swap(m_indices[0], m_indices[2]);
	std::swap(m_indices[0], m_indices[1]);
}

inline void Tripplet::RollRight() throw()
{
	std::swap(m_indices[0], m_indices[2]);
	std::swap(m_indices[1], m_indices[2]);
}

/** The first tripplet index is put in position index0,
* the second tripplet index is put in position index1,
* and the remaining index is put in the yet unused position.
*/
inline void Tripplet::Permute(USHORT index0, USHORT index1) throw()
{
	assert(index0 < 3 && index1 < 3 && index0 != index1);
	Tripplet newTripplet(m_indices[index0], m_indices[index1], m_indices[3-(index0+index1)]);
	*this = newTripplet;
}


inline void  Tripplet::SetFirstIndex(ULONG val) throw()
{
	m_indices[0] = val;
}

inline void Tripplet::SetSecondIndex(ULONG val) throw()
{
	m_indices[1] = val;
}

inline void Tripplet::SetThirdIndex(ULONG val) throw()
{
	m_indices[2] = val;
}

inline void Tripplet::SetIndex(USHORT i, ULONG val) throw()
{
	assert(i<3);
	m_indices[i] = val;
}

inline void Tripplet::AddOffset(ULONG offset) throw()
{
	m_indices[0] += offset;
	m_indices[1] += offset;
	m_indices[2] += offset;
}

// Returns the order index of val if val
// is present in the tripplet, or 4 otherwise.
inline USHORT Tripplet::FindIndex(ULONG val) const throw()
{
	if (m_indices[0] == val)
		return 0;
	if (m_indices[1] == val)
		return 1;
	if (m_indices[2] == val)
		return 2;

	return 4;
}


/** Assumes the tripplets represent neighbouring triangles,
* meaning that the disjunction of the two indices sets is of size 2.
* Returns true if the two triangles are differently oriented
*/
inline bool Tripplet::AreReversed(const Tripplet & lhs, const Tripplet & rhs) throw() {
	USHORT lind0, lind1, rind0, rind1;
	bool lforward, rforward;

	if (rhs.FindIndex(lhs.GetFirstIndex()) > 2) {
		lind0 = 1;
		lind1 = 2;
		lforward = true;
	}
	else if (rhs.FindIndex(lhs.GetSecondIndex()) > 2) {
		lind0 = 0;
		lind1 = 2;
		lforward = false;
	}
	else if (rhs.FindIndex(lhs.GetThirdIndex()) > 2){
		lind0 = 0;
		lind1 = 1;
		lforward = true;
	}
	else {
		return false;
	}

	rind0 = rhs.FindIndex(lhs.GetIndex(lind0));
	rind1 = rhs.FindIndex(lhs.GetIndex(lind1));

	rforward = false;
	if ( (rind0 == 0 && rind1 == 1) ||
		 (rind0 == 1 && rind1 == 2) ||
		  (rind0 == 2 && rind1 == 0) )
		rforward = true;

	return (lforward == rforward);
}

#endif // !defined(AFX_TRIPPLET_H__9283BB44_5348_11D4_AE4D_00609716AB58__INCLUDED_)
